module CeTA where

import Data.List
import Term
import Rule
import TRS
import PCP
import Text.XML.Light 
import Text.XML.Light.Types

type Attribute = (String, String)

data XML =
    Tag String [Attribute] [XML]
  | Text String
  | Element Text.XML.Light.Element
  | Int Int

type Labeling a = [(Rule, a)]

type Diagram = (Peak, [Term])

type CritPairInfo = (Term, [Position], Term, Term, Maybe (Int, Int), [Term])

data CRProof = 
    EmptyTRS
  | ReductionMethod {
      _trsR :: TRS,
      _trsC :: TRS,
      _diagrams :: [Diagram],
      _crProofForC :: CRProof 
    }
  | CompositionalPcpRuleLabeling {
      _phi :: Labeling Int,
      _psi :: Labeling Int,
      _joinsRS :: [CritPairInfo],
      _joinsSR :: [CritPairInfo],
      _trsC :: TRS,
      _crProofForC :: CRProof
    }
  | CompositionalPcps {
      _trsR :: TRS, 
      _trsC :: TRS,
      _trsP :: TRS,
      _joinSequencesForC :: [CritPairInfo],
      _joinSequencesForR :: [CritPairInfo],
      _relativeTerminationProof :: String,
      _crProofForC :: CRProof 
    }

type Fork = (Term, Rule, Position, Term, Rule, Term)

data CRDisproof = CapNotUnif Fork

data Proof = CRProof CRProof | CRDisproof CRDisproof

type CriterionCert = TRS -> IO (Maybe Proof)

type Certificate = (TRS, Proof)


-- Labeling join sequences.

lookupBy :: (a -> a -> Bool) -> a -> [(a,b)] -> Maybe b
lookupBy _ _ [] = Nothing
lookupBy p x ((y, z) : a)
  | p x y     = Just z
  | otherwise = lookupBy p x a

labelOfRule :: Labeling a -> Rule -> a
labelOfRule phi rule
  | Just e <- lookupBy Rule.variant rule phi = e
  | otherwise = error "labelOfRule"

labelsOfRules:: Labeling a -> [Rule] -> [a]
labelsOfRules phi rules = [ labelOfRule phi rule | rule <- rules ]

-- XMLs

text :: String -> CData
text s = CData {cdVerbatim = CDataText, cdData = s, cdLine = Nothing}

int :: Int -> CData
int n = text (show n)

qname :: String -> String -> QName
qname prefix name = QName {qPrefix = Just prefix, qName = name, qURI = Nothing}

attr :: QName -> String -> Attr
attr key val = Attr {attrKey = key, attrVal = val}

tagFuncSym :: (String, Int) -> Element
tagFuncSym (f, n) = 
  unode "funcsym" [unode "name" (text f), unode "arity" (int n)]

tagSignature :: Signature -> Element
tagSignature sig =
  unode "signature" [ tagFuncSym fn | fn <- sig ]

tagTerm :: Term -> Element
tagTerm (V x)    = unode "var" (text x)
tagTerm (F f ts) = 
  unode "funapp" 
    (unode "name" (text f) :
     [ unode "arg" [tagTerm t] | t <- ts ])

instance Node Term where
  node q t = node q (tagTerm t)

tagRule :: Rule -> Element
tagRule (l, r) = unode "rule" [unode "lhs" l, unode "rhs" r]

tagRelrules :: TRS -> Element
tagRelrules trs = 
  unode "relrules" [ tagRule lr | lr <- trs ]

tagRules :: (TRS, TRS) -> Element
tagRules (rs, []) = 
  unode "rules" [ tagRule lr | lr <- rs ]
tagRules (rs, ss) =
  unode "rules" ([ tagRule lr | lr <- rs ] ++ [tagRelrules ss])

tagRelativeTRS :: (TRS, TRS) -> Element
tagRelativeTRS (rs, ss) = 
  unode "trs" [tagRules (rs, ss), 
               tagSignature (TRS.signatureOf (rs ++ ss))]

tagProblem :: (TRS, TRS) -> Element
tagProblem (rs, ss) = 
  add_attrs
    [attr (qname "xmlns" "xsi") "http://www.w3.org/2001/XMLSchema-instance",
     attr (unqual "type") "termination",
     attr (qname "xsi" "noNamespaceSchemaLocation") "/xtc.xsd"]
  (unode "problem" 
     [tagRelativeTRS (rs, ss),
      unode "strategy" (text "FULL")])

showProblem :: (TRS, TRS) -> String
showProblem (rs, ss) = 
  "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n" ++
  showElement (tagProblem (rs, ss)) ++ "\n"

tagCpfVersion :: String -> Element
tagCpfVersion s = unode "cpfVersion" (text s)

tagRuleLabelingFunctionEntry :: (Rule, Int) -> Element
tagRuleLabelingFunctionEntry (rule, n) =
  unode "ruleLabelingFunctionEntry"
      [tagRule rule, unode "label" (int n)]

tagRuleLabelingFunction :: Labeling Int -> Element
tagRuleLabelingFunction phi =
  unode "ruleLabelingFunction"
    [ tagRuleLabelingFunctionEntry (rule, n) | (rule, n) <- phi ]

tagPosition :: Int -> Element
tagPosition i = unode "position" (int (i + 1))

tagPositionInTerm :: Position -> Element
tagPositionInTerm p = unode "positionInTerm" [ tagPosition i | i <- p ]

tagLabels :: Int -> Int -> Element
tagLabels maxLeft right =
  unode "labels" 
    [unode "maxLeft" (int maxLeft),
     unode "right" (int right)]

tagIntermediateTerms :: [Term] -> Element
tagIntermediateTerms ts =
  unode "intermediateTerms" [ tagTerm t | t <- ts ]

tagCritPairInfo :: CritPairInfo -> Element
tagCritPairInfo (t, ps, s, u, labels, vs) = 
  unode "critPairInfo" 
    ([unode "left" t, 
      unode "peak" s, 
      unode "right" u,
      unode "overlapPositions" [ tagPositionInTerm p | p <- ps ]] ++
     f labels ++
     [tagIntermediateTerms vs])
  where
    f Nothing = []
    f (Just (maxLeft, right)) = [tagLabels maxLeft right]

tagJoinSequences :: [CritPairInfo] -> Element
tagJoinSequences joinSequences =
  unode "joinSequences" 
    [ tagCritPairInfo critPairInfo | critPairInfo <- joinSequences ]

tagJoins :: String -> [CritPairInfo] -> Element
tagJoins s diagrams = 
  unode s (tagJoinSequences diagrams)

tagJoinsRS :: [CritPairInfo] -> Element
tagJoinsRS diagrams = tagJoins "joinsRS" diagrams

tagJoinsSR :: [CritPairInfo] -> Element
tagJoinsSR diagrams = tagJoins "joinsSR" diagrams

relativeTerminationProof :: String -> Maybe Element
relativeTerminationProof s = do
  root <- parseXMLDoc s
  proof <- findChild (unqual "proof") root
  findChild (unqual "relativeTerminationProof") proof

tagRelativeTerminationProof :: String -> Element
tagRelativeTerminationProof proof =
  case relativeTerminationProof proof of
    Nothing -> error ("tagRelativeTerminationProof: Failed to parse\n" ++ proof)
    Just e  -> e

tagTRS :: TRS -> Element
tagTRS trs = unode "trs" (tagRules (trs, []))

tagInput :: TRS -> Element
tagInput trs =
  unode "input" (unode "trsInput" (tagTRS trs))

tagTrsC :: TRS -> Element
tagTrsC cs = unode "trsC" (tagRules (cs, []))

tagTrsP :: TRS -> Element
tagTrsP ps = unode "trsP" (tagRules (ps, []))


labelOfRules :: Labeling Int -> TRS -> Int
labelOfRules phi trs =
  maximum (0 : [ i | rule <- trs, Just i <- [lookupBy Rule.variant rule phi] ])

diagramToCritPairInfo :: Labeling Int -> Diagram -> CritPairInfo
diagramToCritPairInfo phi ((t, rulesL, ps, _, s, ruleR, u), vs) =
  (t, ps, s, u, Just (m, n), vs)
  where
    m = labelOfRules phi rulesL
    n = labelOfRules phi [ruleR]

diagramsToJoinsRS :: Labeling Int -> [Diagram] -> [CritPairInfo]
diagramsToJoinsRS phi diagrams =
  [ diagramToCritPairInfo phi diagram | diagram <- diagrams ]

tagCrProof :: CRProof -> Element
tagCrProof crProof = unode "crProof" crProof

instance Node CRProof where
  node q EmptyTRS = 
    node q (unode "orthogonal" ())
  node q r@(ReductionMethod {}) =
    node q $
    unode "compositionalPcpRuleLabeling"
      [tagRuleLabelingFunction phi,
      tagJoinsRS (diagramsToJoinsRS phi (_diagrams r)),
      tagTRS (_trsC r),
      tagCrProof (_crProofForC r)]
    where
      label rule = if elem rule (_trsC r) then 0 else 1
      phi = [ (rule, label rule) | rule <- _trsR r ]
  node q r@(CompositionalPcpRuleLabeling {}) =
    node q $
    unode "compositionalPcpRuleLabeling"
      [tagRuleLabelingFunction (_phi r),
       tagRuleLabelingFunction (_psi r),
       tagJoinsRS (_joinsRS r),
       tagJoinsSR (_joinsSR r),
       tagTRS (_trsC r),
       tagCrProof (_crProofForC r)]
  node q r@(CompositionalPcps {}) =
    node q $
    unode "compositionalPcps"
      [tagTrsC (_trsC r),
       tagTrsP (_trsP r),
       tagJoinSequences (_joinSequencesForC r),
       tagJoinSequences (_joinSequencesForR r),
       tagRelativeTerminationProof (_relativeTerminationProof r),
       tagCrProof (_crProofForC r)]

tagRewriteStep :: [(Rule, Position, Term)] -> Element
tagRewriteStep steps =
  unode "rewriteStep"
    [ e | (rule, p, t) <- steps, 
          e <- [tagPositionInTerm p, tagRule rule, tagTerm t] ]

tagRewriteSequence :: (Term,[(Rule,Position,Term)]) -> Element
tagRewriteSequence (s, steps) =
  unode "rewriteSequence"
    [unode "startTerm" s,
     tagRewriteStep steps]

tagNonJoinableFork :: Fork -> Element
tagNonJoinableFork (t, rule1, p, s, rule2, u) =
  unode "nonJoinableFork" 
    [tagRewriteSequence (s, [(rule1, p, t)]),
     tagRewriteSequence (s, [(rule2, [], u)]),
     unode "capNotUnif" ()]

instance Node CRDisproof where
  node q (CapNotUnif fork) = 
    node q (tagNonJoinableFork fork)

tagCrDisproof :: CRDisproof -> Element
tagCrDisproof crDisproof = unode "crDisproof" crDisproof

instance Node Proof where
  node q (CRProof    crProof)    = node q (tagCrProof    crProof)
  node q (CRDisproof crDisproof) = node q (tagCrDisproof crDisproof)

tagProof :: Proof -> Element
tagProof proof = unode "proof" proof

tagCertificationProblem :: Certificate -> Element
tagCertificationProblem (trs, proof) = 
  add_attrs 
    [attr (qname "xmlns" "xsi") "http://www.w3.org/2001/XMLSchema-instance", 
     attr (qname "xsi" "noNamespaceSchemaLocation") "cpf.xsd"]
    (unode "certificationProblem"
       [tagInput trs,
        unode "cpfVersion" (text "2.1"),
        tagProof proof])

showCertificate :: Certificate -> String
showCertificate certificate =
  "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n" ++
  "<?xml-stylesheet type=\"text/xsl\" href=\"cpfHTML.xsl\"?>\n" ++
  showElement (tagCertificationProblem certificate) ++ "\n"
