module PCPSCert where

import Data.List
import Term
import Rule
import TRS
import SN
import PCP
import Result
import CeTA
import Parser

-- { [t_i, t_{i-1},...,t1,s]) | s -> t1 -> .. -> ti for some i <= k }
at_most_k_steps :: Int -> TRS -> Term -> [[Term]]
at_most_k_steps 0 _   s = [ [s] ]
at_most_k_steps k trs s =
  nubBy eq (tss ++ [ u : ts | ts@(t : _) <- tss, u <- reducts trs t ])
  where 
    eq (v : _) (w : _) = v == w
    tss = at_most_k_steps (k-1) trs s

joinSequence :: Int -> TRS -> Term -> Term -> Maybe [Term]
joinSequence k trs t u
  | t == u    = Just []
  | otherwise =
    case vss of
      vs : _ -> Just (init (tail vs))
      []     -> Nothing
    where
      tss = at_most_k_steps k trs t
      uss = at_most_k_steps k trs u
      vss = [ reverse ts ++ (v : us) | v : ts <- tss, v' : us <- uss, v == v' ]

joinable :: Int -> TRS -> Term -> Term -> Bool
joinable k trs t u = joinSequence k trs t u /= Nothing

pcpIsJoinable :: Int -> TRS -> Bool
pcpIsJoinable k trs = all p (pcpeak trs)
  where p (t,_,_,_,_,_,u) = joinable k trs t u

diagrams :: Int -> TRS -> [Peak] -> Maybe [CritPairInfo]
diagrams _ _    [] = Just []
diagrams k trs ((t, _, ps, xs, s, _, u) : peaks) = do
    vs <- joinSequence k trs t u
    ds <- diagrams k trs peaks
    return ((t, ps, s, u, Nothing, vs) : ds)

pcps' :: Int -> TRS -> [Peak] -> ([CritPairInfo], TRS)
pcps' _ trsC [] = ([], [])
pcps' k trsC ((t, rulesL, ps, xs, s, ruleR, u) : peaks)
  | subset (ruleR : rulesL) trsC       =
      ((t, ps, s, u, Nothing, [s]) : joinSequences, trsP)
  | Just vs <- joinSequence k trsC t u =
      ((t, ps, s, u, Nothing, vs)  : joinSequences, trsP)
  | otherwise                          = 
      (joinSequences, (s, t) : (s, u) : trsP)
  where
    (joinSequences, trsP) = pcps' k trsC peaks

-- PCPS(R,C)
-- Instead of R, parallel critical peaks of R must be given. 
pcps :: Int -> [Peak] -> TRS -> ([CritPairInfo], TRS)
pcps k peaks trsC = (joinSequences, nub trsP)
  where (joinSequences, trsP) = pcps' k trsC peaks

-- This is the simulation of HM11 by SH22.
-- Due to the use of PCP(R) it is less efficient.
hm11 :: RTermination -> Int -> CriterionCert
hm11 proveSN k [] = return (Just (CRProof EmptyTRS))
hm11 proveSN k trsR
  | TRS.leftLinear trsR,
    Just joinSequencesForR <- diagrams k trsR peaks = do
      m <- proveSN trsP trsR
      case m of
        MAYBE -> return Nothing
        NO _  -> return Nothing
        YES relativeTerminationProof ->
          let crProof = CompositionalPcps {
              _trsR = trsR,
              _trsC = [],
              _trsP = trsP,
              _joinSequencesForC = joinSequencesForC,
              _joinSequencesForR = joinSequencesForR,
              _relativeTerminationProof = relativeTerminationProof,
              _crProofForC = CeTA.EmptyTRS } 
            in 
              return (Just (CRProof crProof))
  | otherwise = return Nothing
  where
    peaks = pcpeak trsR
    (joinSequencesForC, trsP) = pcps k peaks []

power :: [a] -> [[a]]
power [] = [ [] ]
power (x : xs) = yss ++ [ x : ys | ys <- yss ]
  where yss = power xs

pcpsCriterion' :: RTermination -> Int -> CriterionCert -> TRS -> [TRS] -> [Peak] -> [CritPairInfo] -> IO (Maybe Proof)
pcpsCriterion' _  _ _ _    []             _     _                  =
  return Nothing
pcpsCriterion' tt k cr trsR (trsC : trsCs) peaks joinSequencesForR = do
  m <- tt trsP trsR
  case m of
    YES relativeTerminationProof -> do
      v <- cr trsC
      case v of
        Nothing             -> pcpsCriterion' tt k cr trsR trsCs peaks joinSequencesForR
        Just (CRDisproof _) -> pcpsCriterion' tt k cr trsR trsCs peaks joinSequencesForR
        Just (CRProof crProofForC) ->
          let crProof = CompositionalPcps {
            _trsR = trsR,
            _trsC = trsC,
            _trsP = trsP,
            _joinSequencesForC = joinSequencesForC,
            _joinSequencesForR = joinSequencesForR,
            _relativeTerminationProof = relativeTerminationProof,
            _crProofForC = crProofForC
          } 
          in
            return (Just (CRProof crProof))
    _  -> pcpsCriterion' tt k cr trsR trsCs peaks joinSequencesForR
  where
    (joinSequencesForC, trsP) = pcps k peaks trsC

pcpsCriterion :: RTermination -> Int -> Bool -> CriterionCert -> CriterionCert
pcpsCriterion tt k _        cr []   = return (Just (CRProof EmptyTRS))
pcpsCriterion tt k includeR cr trsR
  | TRS.leftLinear trsR,
    Just joinSequences <- diagrams k trsR peaks =
      pcpsCriterion' tt k cr trsR trsCs peaks joinSequences
  | otherwise = return Nothing
  where
    peaks = pcpeak trsR
    n = length trsR
    trsCs' = [ xs | xs <- power trsR, length xs /= n ]
    trsCs = if includeR then trsR : trsCs' else trsCs'

pcpsPlus :: RTermination -> Int -> CriterionCert
pcpsPlus tt k trsR = pcpsCriterion tt k False (pcpsPlus tt k) trsR
