-- Certifiable rule labeling based on parallel critical pairs
module PRLCert where

import Data.List
import Term
import Rule
import TRS
import Rewriting
import PCP
import SMT
import CeTA

-- s ->_{l -> r} ...
type Step = (Rule, Term)

-- s1 ->_{l1 -> r1} s2 ->_{l2 -> r2} ... sn ->_{ln -> rn} t
type Sequence = ([Step], Term)

-- s ->* -||-> ->* t
type ParallelSequence = (Sequence, [Rule], Term, Sequence)

-- t ->* -||-> ->* *<- <-||- *<- u
type JoinParallelSequence = (ParallelSequence, ParallelSequence)

data PeakType = PhiPsi | PsiPhi deriving Show

-- t <-||- -> u and t ->* -||-> ->* *<- <-||- *<- u
type CandidateDiagrams = (Peak, [JoinParallelSequence])

-- Generating diagrams.

geqEmb :: Eq a => [a] -> [a] -> Bool
geqEmb _ []       = True 
geqEmb [] (_ : _) = False
geqEmb (x : xs) (y : ys)
  | x == y    = geqEmb xs ys
  | otherwise = geqEmb xs (y : ys)

rulesOfSteps :: [Step] -> [Rule]
rulesOfSteps steps = [ rule | (rule, _) <- steps ]

geqSequence :: Sequence -> Sequence -> Bool
geqSequence (steps1, s) (steps2, t) =
  s == t && geqEmb (rulesOfSteps steps1) (rulesOfSteps steps2)

minimalSequences :: [Sequence] -> [Sequence]
minimalSequences [] = []
minimalSequences (s : ss)
  | any (geqSequence s) ss = minimalSequences ss
  | otherwise              = 
      s : minimalSequences [ s' | s' <- ss, not (geqSequence s' s) ]

reverseSequences :: Int -> TRS -> Term -> [Sequence]
reverseSequences 0 _   s = [([], s)]
reverseSequences n trs s = 
  minimalSequences
    (([], s) :
     [ ((rule, u) : steps, u)
     | (steps, t) <- reverseSequences (n - 1) trs s,
       notElem s [ t' | (_, t') <- steps ],
       rule <- trs,
       u <- reducts [rule] t,
       s /= u ])

parallelSequences :: Int -> TRS -> [String] -> Term -> [ParallelSequence]
parallelSequences n trs xs s =
  [ (([], s), [], s, seq3) | seq3 <- ss ] ++
  [ (seq1, rules2, u, seq3)
  | seq1@(steps1, t) <- ss,
    length steps1 < n,
    (rules2, u) <- restrictedParallelStep trs xs t,
    rules2 /= [] && not (length rules2 == 1 && subset rules2 (rulesOfSteps steps1)),
    seq3 <- reverseSequences (n - length steps1 - 1) trs u ]
  where 
    ss = reverseSequences n trs s

-- joinSequence n trs s t
joinParallelSequences :: Int -> TRS -> Term -> [String] -> Term -> [JoinParallelSequence]
joinParallelSequences n trs s xs t =
   [ (psR, psL)
   | psR@(_, _, _, (_, u)) <- parallelSequences n trs (Term.variables s) s,
     psL@(_, _, _, (_, v)) <- parallelSequences n trs xs t,
     u == v ]

candidateDiagrams :: TRS -> TRS -> Int -> [CandidateDiagrams]
candidateDiagrams rs cs k =
  [ ((t, rulesL, ps, xs, s, ruleR, u), joinParallelSequences k rs t xs u)
  | (t, rulesL, ps, xs, s, ruleR,  u) <- pcpeak rs,
    -- discard parallel critical peaks of C-steps
    not (TRS.subsume cs (ruleR : rulesL)) ]

-- Labels

labelsOfSequence:: Labeling a -> Sequence -> [a]
labelsOfSequence phi (steps, _) = [ labelOfRule phi rule | (rule, _) <- steps ]


-- Encoding functions.

labeling :: String -> TRS -> Labeling Exp
labeling x trs =
  [ (rule, Var (x ++ show i)) | (rule, i) <- zip trs [0 :: Integer ..] ]

varDiagram :: PeakType -> Int -> Exp
varDiagram PhiPsi i = Var ("phi_psi_" ++ show i) 
varDiagram PsiPhi i = Var ("psi_phi_" ++ show i) 

gt2 :: [Exp] -> [Exp] -> Formula
gt2 ks ms = conj [ disj [ Gt k m | k <- ks ] | m <- ms ]

geq2 :: [Exp] -> [Exp] -> Formula
geq2 ks ms = conj [ disj [ Geq k m | k <- nub ks ] | m <- nub ms ]

-- ...L stands for "<-"
-- ...R stands for "->"
decreasingness :: Labeling Exp -> Labeling Exp -> [CandidateDiagrams] -> Formula
decreasingness phi psi diagrams =
  conj [ disj [ conj [Eq (varDiagram peakType i) (SMT.Val j),
                      gt2  ks  (labelsOfSequence psi' stepsR1),
                      geq2 ms  (labelsOfRules    psi' rulesR2),
                      gt2  kms (labelsOfSequence psi' stepsR3),
                      gt2  kms (labelsOfSequence phi' stepsL3),
                      geq2 ks  (labelsOfRules    phi' rulesL2),
                      gt2  ms  (labelsOfSequence phi' stepsL1)]
              | (j, ((stepsR1, rulesR2, _, stepsR3), 
                (stepsL1, rulesL2, _, stepsL3))) <- zip [0..] jss ]
       | (i, ((_, rulesL, _, _, _, ruleR, _), jss)) <- zip [0..] diagrams,
         (peakType, phi', psi') <- [(PhiPsi, phi, psi), (PsiPhi, psi, phi)],
         let ks = labelsOfRules phi' rulesL,
         let ms = [labelOfRule psi' ruleR],
         let kms = nub (ks ++ ms) ]

bottomConstraint :: Labeling Exp -> TRS -> TRS -> Formula
bottomConstraint phi rs cs =
  conj ([ Eq (labelOfRule phi rule) (Val 0) | rule <- cs  ] ++
        [ Gt (labelOfRule phi rule) (Val 0) | rule <- diff rs cs ])

constraint :: TRS -> TRS -> Labeling Exp -> Labeling Exp -> [CandidateDiagrams] -> Formula
constraint rs cs phi psi ds =
  conj [bottomConstraint phi rs cs,
        bottomConstraint psi rs cs,
        decreasingness phi psi ds]
        
evalLabeling :: Model -> Labeling Exp -> Labeling Int
evalLabeling model phi =
  [ (rule, evalExp model e) | (rule, e) <- phi ]

compact :: Eq a => [a] -> [a]
compact [] = []
compact [x] = [x]
compact (x : y : zs)
  | x == y    =     compact (y : zs)
  | otherwise = x : compact (y : zs) 

termsInParallelSequence :: ParallelSequence -> [Term]
termsInParallelSequence ((steps1, _), _, t, (steps2, _)) =
  reverse [ s | (_, s) <- steps1 ] ++ t :
  reverse [ u | (_, u) <- steps2 ]

intermediates:: JoinParallelSequence -> [Term]
intermediates (seqR, seqL) = compact (ts ++ reverse us) 
  where
    ts = termsInParallelSequence seqR
    us = termsInParallelSequence seqL


decodeDiagrams :: PeakType -> Labeling Int -> Labeling Int -> [CandidateDiagrams] -> Model -> [CritPairInfo]
decodeDiagrams peakType phi' psi' ds model =
  [ (t, ps, s, u, Just (maxLeft, right), intermediates js) 
  | (i, ((t, rulesL, ps, _, s, ruleR, u), jss)) <- zip [0..] ds,
    let js = jss !! evalExp model (varDiagram peakType i),
    let maxLeft = maximum (0 : labelsOfRules phi' rulesL),
    let right = labelOfRule psi' ruleR
  ]


power :: [a] -> [[a]]
power [] = [ [] ]
power (x : xs) = yss ++ [ x : ys | ys <- yss ]
  where yss = power xs

compositionalPRLLoop2 :: String -> [Int] -> CriterionCert -> TRS -> TRS -> Labeling Exp -> Labeling Exp -> IO (Maybe Proof)
compositionalPRLLoop2 _   []       _  _    _    _   _   = return Nothing
compositionalPRLLoop2 smt (k : ks) cr trsR trsC phi psi = do
  m <- sat smt (constraint trsR trsC phi psi diagrams)  
  case m of
    Nothing    -> compositionalPRLLoop2 smt ks cr trsR trsC phi psi
    Just model -> do
      m <- cr trsC
      case m of
        Nothing -> 
          compositionalPRLLoop2 smt ks cr trsR trsC phi psi
        Just (CRDisproof crProofForC) ->
          compositionalPRLLoop2 smt ks cr trsR trsC phi psi
        Just (CRProof crProofForC) ->
          let phi' = evalLabeling model phi in
          let psi' = evalLabeling model psi in
          let proof = CompositionalPcpRuleLabeling {
            _phi = phi',
            _psi = psi',
            _joinsRS = decodeDiagrams PhiPsi phi' psi' diagrams model,
            _joinsSR = decodeDiagrams PsiPhi psi' phi' diagrams model,
            _trsC = trsC,
            _crProofForC = crProofForC
          }
          in
            return (Just (CRProof proof))
  where
    diagrams = candidateDiagrams trsR trsC k

compositionalPRLLoop1 :: String -> [Int] -> CriterionCert -> TRS -> [TRS] -> Labeling Exp -> Labeling Exp -> IO (Maybe Proof)
compositionalPRLLoop1 _   _  _  _    []             _   _   = return Nothing
compositionalPRLLoop1 smt ks cr trsR (trsC : trsCs) phi psi = do
  m <- compositionalPRLLoop2 smt ks cr trsR trsC phi psi
  case m of
    Nothing -> compositionalPRLLoop1 smt ks cr trsR trsCs phi psi
    m       -> return m

compositionalPRL' :: String -> [Int] -> CriterionCert -> TRS -> [TRS] -> IO (Maybe Proof)
compositionalPRL' _   _  _  []   _     = return (Just (CRProof EmptyTRS))
compositionalPRL' smt ks cr trsR trsCs
  | TRS.leftLinear trsR = compositionalPRLLoop1 smt ks cr trsR trsCs phi psi
  | otherwise           = return Nothing
  where
    phi = labeling "phi_" trsR
    psi = labeling "psi_" trsR

emptiness :: CriterionCert
emptiness [] = return (Just (CRProof EmptyTRS))
emptiness _  = return Nothing

prl :: String -> [Int] -> CriterionCert
prl smt ks trsR = compositionalPRL' smt ks emptiness trsR [[]]

compositionalPRL :: String -> [Int] -> CriterionCert -> CriterionCert
compositionalPRL smt ks cr trsR = do
  compositionalPRL' smt ks cr trsR trsCs
  where 
    n = length trsR
    trsCs =
      [] : trsR : 
      [ trsC | trsC@(_ : _) <- power trsR, length trsC /= n ]
