module SN where

import Data.List
import System.IO
import System.Process
import Term
import TRS
import Result
import qualified WST
import qualified CeTA
import Text.XML.Light

data Format = WST | XML

type Tool = (Format, String)

type RTermination = TRS -> TRS -> IO Result


-- A term t is non-terminating with respect to R/S if
-- R contains rule of form l -> C[l sigma] such that
-- x in Var(x sigma) cap Var(C) for some variable x.
-- See Lemma 10 of Hirokawa and Middeldorp (JAR 2011).

surroundingVariables :: Term -> Position -> [String]
surroundingVariables _        []      = []
surroundingVariables (F _ ts) (i : p) = 
  nub [ x | (j, tj) <- zip [0..] ts, 
            x <- if i == j then surroundingVariables tj p else Term.variables tj ]
surroundingVariables _ _ = error "surroundingVariables"

loopInstance :: (Term, Term, Position) -> Bool
loopInstance (l, r, p) =
  case match l (subtermAt r p) of
    Nothing -> False
    Just sigma -> any check (surroundingVariables r p)
      where
        check x = elem x (Term.variables (Term.substitute (V x) sigma)) 

-- True means that R/S is non-terminating. 
-- False means that we do not know.
looping :: TRS -> TRS -> Bool
looping [] _  = False
looping rs ss =
  any (\(l, r) -> Term.subsume l r) rs ||
  any loopInstance [ (l, r, p) | (l, r) <- ss, p <- positions r ]

showProblem :: (Format, TRS, TRS) -> String
showProblem (WST, rs, ss) = WST.showProblem (rs, ss)
showProblem (XML, rs, ss) = CeTA.showProblem (rs, ss)

execute :: String -> String -> IO (String, String)
execute tool input = do
  (Just hin, Just hout, Just herr, _) <-
    createProcess (proc tool ["/dev/stdin"]) {
      std_in = CreatePipe,
      std_out = CreatePipe,
      std_err = CreatePipe }
  hPutStr hin input
  hClose hin
  errMsg <- hGetContents herr
  msg <- hGetContents hout
  return (msg, errMsg)

writeError :: a -> String -> IO a
writeError x s = do
  hPutStrLn stderr s
  return x

run :: Tool -> TRS -> TRS -> IO Result
run (format, tool) rs ss = do
  -- putStrLn ("-- input --\n" ++ showProblem (format, rs, ss) ++ "\n")
  (s,e) <- execute tool (showProblem (format, rs, ss))
  -- putStrLn ("-- output --\n" ++ s ++ "\n")
  v <- firstLine (lines s) e
  case v of
    "YES"   -> return (YES tool)
    "NO"    -> return (NO tool)
    "MAYBE" -> return MAYBE
    v       -> error ("SN.run: " ++ v)
  where
    firstLine []     e = writeError "" (unlines [tool ++ " failed:", e])
    firstLine ("":_) e = writeError "" (unlines [tool ++ " failed:", e])
    firstLine (s:_)  _ = return (head (words s))

-- check termination of R/S by command tool
sn :: Tool -> RTermination
sn tool rs ss
  | looping rs ss = return (NO "R/S loops")
  | otherwise     = run tool rs ss

-- A parsing function for outputs of termination tools.
readOutput :: String -> (String, String)
readOutput s =
  case lines s of
    []           -> ("", "")
    line : lines -> 
      case words line of
        []       -> ("",   unlines lines)
        word : _ -> (word, unlines lines)

runCert :: Tool -> TRS -> TRS -> IO Result
runCert (format, tool) rs ss = do
  -- putStrLn ("-- input --\n" ++ showProblem (format, rs, ss) ++ "\n")
  (s,e) <- execute tool (showProblem (format, rs, ss))
  -- putStrLn ("-- output --\n" ++ s ++ "\n")
  case readOutput s of
    ("YES", proof) -> return (YES proof)
    ("NO", proof)  -> return (NO proof)
    ("MAYBE", _)   -> return MAYBE
    _              -> error (unlines [tool ++ " failed:", e])

snCert :: Tool -> TRS -> TRS -> IO Result
snCert tool rs ss
  | looping rs ss = return (NO "R/S loops")
  | otherwise     = runCert tool rs ss
